%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% sentences %%%

:- [input_ok].
:- [nowords].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% lexicon %%%

:- op(600,xfy, ':').

:- [lexicon].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% rules %%%

:- [rules].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% grammar model %%%

:- [gm].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% feature structure preprocessor 

g_schema(subcat:X,   [X,_,_,_,_,_,_,_,_,_,_,_,_,_]). %%% main struct
g_schema(case:X,     [_,X,_,_,_,_,_,_,_,_,_,_,_,_]).
g_schema(person:X,   [_,_,X,_,_,_,_,_,_,_,_,_,_,_]).
g_schema(agr:X,      [_,_,_,X,_,_,_,_,_,_,_,_,_,_]).
g_schema(pos:X,      [_,_,_,_,X,_,_,_,_,_,_,_,_,_]).
g_schema(vtyp:X,     [_,_,_,_,_,X,_,_,_,_,_,_,_,_]).
g_schema(det:X,      [_,_,_,_,_,_,X,_,_,_,_,_,_,_]).
g_schema(def:X,      [_,_,_,_,_,_,_,X,_,_,_,_,_,_]).
g_schema(rreq:X,     [_,_,_,_,_,_,_,_,X,_,_,_,_,_]).
g_schema(ragr:X,     [_,_,_,_,_,_,_,_,_,X,_,_,_,_]).
g_schema(tagr:X,     [_,_,_,_,_,_,_,_,_,_,X,_,_,_]).
g_schema(vagr:X,     [_,_,_,_,_,_,_,_,_,_,_,X,_,_]).
g_schema(semcat:X,   [_,_,_,_,_,_,_,_,_,_,_,_,X,_]).
g_schema(sem:X,      [_,_,_,_,_,_,_,_,_,_,_,_,_,X]).

g_schema(number:X,  [X,_]).                %%% agreement structure
g_schema(gender:X,  [_,X]).
 
g_schema(day:X,   [X,_,_]).           %%% time agreement structure
g_schema(tod:X,   [_,X,_]).
g_schema(time:X,  [_,_,X]).

g_schema(subj:X,  [X,_,_,_,_,_]).     %%% verb agreement structure
g_schema(obj1:X,  [_,X,_,_,_,_]).
g_schema(obj2:X,  [_,_,X,_,_,_]).
g_schema(vcomp:X, [_,_,_,X,_,_]).
g_schema(scomp:X, [_,_,_,_,X,_]).
g_schema(ppobj:X, [_,_,_,_,_,X]).

g_schema(req:X,   [X]).      %%% relative clause request structure

match([Kat,[A,B,C,D,E,F,U,V,W1,X,H1,I1,G,J]],
      [Kat,[A,B,C,D,E,F,U,V,W2,X,H2,I2,G,J]],
      [Kat,[A,B,C,D,E,F,U,V,WR,X,HR,IR,G,J]]) :- 
  match_plus(H1,H2,HR),
  match_plus(W1,W2,WR),
  match_plus(I1,I2,IR).  

match_plus(A,B,R) :-
  (var(A);var(B)),!,
  R = A,
  R = B.

match_plus(A,complete,R) :- 
  proper(A),
  R = complete.

match_plus(complete,B,R) :- 
  proper(B),
  R = complete.

match_plus(A,B,R) :-
  match_plus_list(A,B,R).

match_plus_list([],[],[]).

match_plus_list([A1|A],[B1|B],[R1|R]) :-
  match_item(A1,B1,R1),
  match_plus_list(A,B,R).

match_item(A,B,R) :-
  (var(A);var(B)),!,
  A = R,
  B = R.

match_item([need,Type],[has,Type],[satisfied,Type]).
match_item([has,Type],[need,Type],[satisfied,Type]).
match_item([has,Type],[may,Type],[satisfied,Type]).
match_item([may,Type],[has,Type],[satisfied,Type]).

proper([]).

proper([A1|A]) :- var(A1),!,proper(A).
proper([[may,_]|A]) :- proper(A).
proper([[satisfied,_]|A]) :- proper(A).

compute_new([_,[_,_,_,_,_,_,_,_,W1,_,H1,I1,_,_]],
            [_,[_,_,_,_,_,_,_,_,W2,_,H2,I2,_,_]],
            [K,[A,B,C,D,E,F,U,V,WR,X,HR,IR,G,J]],
            [_,[_,_,_,_,_,_,_,_,WW,_,HW,IW,_,_]],
            [K,[A,B,C,D,E,F,U,V,WE,X,HE,IE,G,J]]) :- !,
  compute_new_plus_pre_check(H1,H2,HR,HW,HE),!,
  compute_new_plus_pre_check(W1,W2,WR,WW,WE),!,
  compute_new_plus_pre_check(I1,I2,IR,IW,IE).  

compute_new_plus_pre_check(A,B,C,D,E) :-
  var(A),var(B),var(C),var(D),var(E),!,
  C=E.

compute_new_plus_pre_check(A,B,C,D,E) :-
  compute_new_plus(A,B,C,D,E).

compute_new_plus(complete,_,Neu,_,Neu).
compute_new_plus(_,complete,Neu,_,Neu).

compute_new_plus([],[],[],[],[]).
compute_new_plus([L|LL],[R|RR],[N|NN],[W|WW],[E|EE]) :-
  compute_new_item(L,R,N,W,E),
  compute_new_plus(LL,RR,NN,WW,EE).

compute_new_item(RegelLinks,RegelRechts,RegelNeu,WertLinks,Ergebnis) :-
  RegelLinks == RegelRechts,
  RegelRechts == RegelNeu,!,
  go_for_the_last_test(RegelNeu,WertLinks,Ergebnis).

compute_new_item(_,_,RegelNeu,_,Ergebnis) :-
  Ergebnis = RegelNeu.

go_for_the_last_test(RegelNeu,WertLinks,Ergebnis) :-
  RegelNeu == WertLinks,!,
  Ergebnis = RegelNeu.

go_for_the_last_test(RegelNeu,WertLinks,Ergebnis) :- !,
  match_item(RegelNeu,WertLinks,Ergebnis).

%%% translation of feature structures

g_translate([],_).

g_translate([Kopf|Rest],Liste) :-
  g_trans_feature(Kopf,Liste),
  g_translate(Rest,Liste).

g_trans_feature(Merkmal:Auspraegung,Liste) :-
   var(Auspraegung),!,
   g_schema(Merkmal:Auspraegung,Liste).

g_trans_feature(tagr:initial,Liste) :-
   g_trans_feature(tagr:[time:[may,_],tod:[may,_],day:[may,_]],Liste).

g_trans_feature(Merkmal:Auspraegung,Liste) :-
   atomic(Auspraegung),!,
   g_schema(Merkmal:Auspraegung,Liste).

g_trans_feature(Merkmal:[Zustand|Typ],Liste) :-
   (Zustand=has;Zustand=may;Zustand=need;Zustand=satisfied),!,
   g_schema(Merkmal:[Zustand|Typ],Liste).

g_trans_feature(Merkmal:[Kopf|Auspraegung],Liste) :-
   Kopf = list,!,
   g_schema(Merkmal:Auspraegung,Liste).

g_trans_feature(Merkmal:[Aus|Praegung],Liste) :- !,
   g_translate([Aus|Praegung],AS),
   g_schema(Merkmal:AS,Liste). 

g_trans_feature(Merkmal:Auspraegung,Liste) :-
   g_schema(Merkmal:Auspraegung,Liste).

%%% translation of grammar components 

g_t_grammar([Kat,M],[Kat,MS]) :- g_translate(M,MS), !.

g_t_grammar_rec([H|T],[H1|T1]) :- 
   g_t_grammar(H,H1),
   g_t_grammar_rec(T,T1).

g_t_grammar_rec([],[]).

%%% Lexicon compilation

compile_lexicon :-
   lexicon(Wort,Kat,Merkmale),
   atom(Wort),
   g_translate([sem:[list,Kat,Wort]|Merkmale],MerkmalsStruktur),
   word_number(Nummer),
   assert(wordform(Wort,[Kat,MerkmalsStruktur],Nummer)),
   fail.

compile_lexicon :-
   lexicon([W|R],K,M),
   compile_lexicon_rec([W|R],K,M),
   fail.
 
compile_lexicon.
 
compile_lexicon_rec([],_,_).

compile_lexicon_rec([Wort|AndereWoerter],Kat,Merkmale) :- 
   g_translate([sem:[list,Kat,Wort]|Merkmale],MerkmalsStruktur),
   word_number(Nummer),
   assert(wordform(Wort,[Kat,MerkmalsStruktur],Nummer)),
   compile_lexicon_rec(AndereWoerter,Kat,Merkmale).

init_word_counter :-
  abolish(word_counter,1),
  assert(word_counter(1)).

word_number(X) :-
  word_counter(X),
  abolish(word_counter,1),
  Y is X+1,
  assert(word_counter(Y)).

%%% Grammar compilation

compile_grammar :-
   grammar(X,Y),
   compile_rule(X,Y),
   fail.

compile_grammar. 
 
compile_rule(                                   % unary rule
   [LinksKat,LinksML],[[RechtsKat,RechtsML]]) :-
   g_t_grammar([LinksKat,[sem:[list,Nummer,LinksKat,S1]|LinksML]],
               LinksStruktur),
   g_t_grammar_rec([[RechtsKat,[sem:S1|RechtsML]]],RechtsStruktur),
   rule_number(Nummer),
   assert(rule(LinksStruktur,RechtsStruktur,Nummer)).

compile_rule(                                  % binary rule
           [LinksKat,LinksML],
           [[Rechts1Kat,Rechts1ML],
            [Rechts2Kat,Rechts2ML]]) :-
   g_t_grammar([LinksKat,[sem:[list,Nummer,LinksKat,S1,S2]|LinksML]],
               LinksStruktur),
   g_t_grammar_rec([[Rechts1Kat,[sem:S1|Rechts1ML]],
                    [Rechts2Kat,[sem:S2|Rechts2ML]]],
                   RechtsStruktur),
   rule_number(Nummer),
   assert(rule(LinksStruktur,RechtsStruktur,Nummer)).

init_rule_counter :-
  abolish(rule_counter,1),
  assert(rule_counter(1)).

rule_number(X) :-
  rule_counter(X),
  abolish(rule_counter,1),
  Y is X+1,
  assert(rule_counter(Y)).

learn_rule_counter :-
  rule_counter(X),
  X>500,!,
  abolish(lrule_counter,1),
  assert(lrule_counter(X)).

learn_rule_counter :-
  abolish(rule_counter,1),
  abolish(lrule_counter,1),
  assert(rule_counter(501)),
  assert(lrule_counter(501)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% sentence decomposition %%%

compile_sentence([],_).

compile_sentence([Wort|_],Offset) :-
   OffsetPlusEins is Offset+1,
   wordform(Wort,Struktur,Nummer),
   assert(word(Struktur,Offset,OffsetPlusEins,Nummer)),
   assert(foundword),
   fail.

compile_sentence([Wort|Restsatz],Offset) :- 
   check_foundword(Wort),
   OffsetPlusEins is Offset+1,
   compile_sentence(Restsatz,OffsetPlusEins).

check_foundword(_) :-
   foundword,!,
   abolish(foundword,0).

check_foundword(Wort) :-
   tell(user),
   write('unknown word: '),told,write(Wort),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% report of internal states %%%

%%% report all rules 

report_rules :- 
  init_rule_counter, abolish(rule,3),
  compile_grammar,
  tell('rules.inf'), report_r, told, tell(user).

r :- 
  tell('rules.inf'), report_r, told, tell(user).

report_learned_rules :- 
  tell('rules.learn.inf'), report_lr, told, tell(user).

report_r :-
  rule(Links,Rechts,Nummer),
  write('['), write(Links), write(','),
  write(Rechts), write(','),
  write(Nummer), write(']'), nl,
  fail.

report_r.

report_lr :-
  rule(Links,Rechts,Nummer),
  Nummer>500,
  write('['), write(Links), write(','),
  write(Rechts), write(','),
  write(Nummer), write(']'), nl,
  fail.

report_lr.

%%% output of parses on screen

report_parses :-
   length(Laenge),
   chart([s,Parse],0,Laenge,_,_,_,_),
   write(Parse),nl,
   fail.

report_parses.
 
%%% count the number of parses

parse_count :-
   length(Laenge),
   chart([s,_],0,Laenge,_,_,_,_),
   parse_counter(X),
   abolish(parse_counter,1),
   Y is X+1,
   assert(parse_counter(Y)),
   fail.

parse_count.

%%% construct parse trees from the chart

construct_tree(ChartNr,[Wort,Kat]) :-
   chart(_,_,_,0,_,WortNr,ChartNr),!,
   wordform(Wort,[Kat|_],WortNr).

construct_tree(ChartNr,[RegelNr,Kat,Ast]) :-
   chart([Kat|_],_,_,RegelNr,RechtsNr,0,ChartNr),!,
   construct_tree(RechtsNr,Ast).
   
construct_tree(ChartNr,[RegelNr,Kat,LinkerAst,RechterAst]) :-
   chart([Kat|_],_,_,RegelNr,LinksNr,RechtsNr,ChartNr),
   construct_tree(LinksNr,LinkerAst),
   construct_tree(RechtsNr,RechterAst).
   
%%% check for unknown words in sentences

wordcheck(A,B) :-
   abolish(wordform,3),
   init_word_counter,
   compile_lexicon,
   do_wordcheck(A,B).

do_wordcheck(X,X).  

do_wordcheck(StartSatz,EndSatz) :- 
   input(StartSatz,Satz),!,
   wordck(StartSatz,Satz),
   NaechsterSatz is StartSatz+1,
   do_wordcheck(NaechsterSatz,EndSatz).

do_wordcheck(StartSatz,EndSatz) :- 
   NaechsterSatz is StartSatz+1,
   do_wordcheck(NaechsterSatz,EndSatz).

wordck(_,[]).

wordck(SNr,[Wort|RestSatz]) :-
   wordform(Wort,_,_),!,
   wordck(SNr,RestSatz).

wordck(SNr,[Wort|RestSatz]) :-
   write(SNr),write(':'),write(Wort),nl,
   wordck(SNr,RestSatz).

%%% sort out no-words

filter :-
  abolish(wordform,3),
  init_word_counter,
  compile_lexicon,
  do_filter(499,820),
  told,told.

do_filter(X,X).

do_filter(StartSatz,EndSatz) :-
  input(StartSatz,Satz),!,
  filterchk(StartSatz,Satz),
  NaechsterSatz is StartSatz+1,
  do_filter(NaechsterSatz,EndSatz).

do_filter(StartSatz,EndSatz) :-
  NaechsterSatz is StartSatz+1,
  do_filter(NaechsterSatz,EndSatz).

filterchk(SNr,[]) :-
  tell('input_ok.pl'),
  input(SNr,Satz),
  write(input(SNr,Satz)),write('.'),nl.

filterchk(SNr,[Wort|_]) :-
  noword(Wort),!,
  tell('input_noword.pl'),
  input(SNr,Satz),
  write(input(SNr,Satz)),write('.'),nl.

filterchk(SNr,[_|RestSatz]) :-
  filterchk(SNr,RestSatz).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% bottom-up chart parser %%%

%% these are two frame-routines: initialisation & output

parse(X,Y) :-
   abolish(parse_input,2),
   assert(parse_input(X,Y)),
   parse.

parse_and_report(X) :-
   parse(X),
   report_unlearned_parses(X),
   unix(system(copy_parsed_inf)).

% a single sentence 

parse(X) :- 
   abolish(parse_input,1),
   assert(parse_input(X)),
   tell('chart.log'),
   init,
   process(0),
   told, tell(user),
   report_parses.

% many sentences
 
parse :-
   parse_input(X,Y),
   X is Y.

parse :-
   parse_input(X,_),
   input(X,_),!,
   parse_go.

parse :-
   parse_input(X,Y),
   Xplusone is X+1,
   abolish(parse_input,2),
   assert(parse_input(Xplusone,Y)),
   parse. 
 
parse_go :- 
   parse_input(X,Y),
   Xplusone is X+1,
   abolish(parse_input,1),
   abolish(parse_input,2),
   assert(parse_input(X)),
   assert(parse_input(Xplusone,Y)),
   init,
   tell('chart.log'),
   process(0),
   told, tell(user),
   abolish(parse_counter,1),
   assert(parse_counter(0)),
   parse_count,
   parse_counter(C),
   write(X),write(':'),write(C),nl,
   parse.

% compile lexicon, grammar, and sentence

init :-
   parse_input(SatzNummer),
   input(SatzNummer,Satz),
   write(input(Satz)),nl, 
   abolish(wordform,3),
   abolish(rule,3),
   abolish(word,4),
   abolish(length,1),
   length(Satz,Laenge),
   assert(length(Laenge)),
   init_word_counter,
   init_rule_counter,
   compile_lexicon,
   compile_grammar,
   compile_sentence(Satz,0),
   clear_chart.

%%% this is the main loop

process(Laenge) :- length(Laenge),!.

process(Position) :-
   announce_scanner(Position),
   scanner(Position,NewPosition),
   announce_completer,
   completer(NewPosition),
   process(NewPosition).

announce_scanner(Position) :-
   write('=== At Position '),write(Position),write(' ==='),nl,
   write('Scanner...'),nl.

announce_completer :-
   write('Completer...'),nl.

%% scanner: consume a word from the input

scanner(Position,NeuePosition) :-
   word(Wort,Position,NeuePosition,Nummer),
   store(chart(Wort,Position,NeuePosition,0,0,Nummer,_)),
   fail.

scanner(Position,NeuePosition) :-
   NeuePosition is Position+1.
   
%% completer: consume completed rules from the chart

completer(Position) :-
   chart(Phrase,Von,Position,_,_,_,N),   % find completed rule
   complete(Phrase,Von,Position,N),      % use it to complete more
   fail.

completer(_).

complete(Rechts,MittePos,EndePos,RNr) :-
   rule(Neu,[L,R],RegelNr),         % for any binary rule
   match(R,Rechts,_),               % that matches phrase on right
   chart(Links,AnfangPos,MittePos,_,_,_,LNr),  
   match(L,Links,_),                % that has left half
   compute_new(L,R,Neu,Links,Ergebnis),
   store(chart(Ergebnis,AnfangPos,EndePos,RegelNr,LNr,RNr,NNr)),
   complete(Ergebnis,AnfangPos,EndePos,NNr),
   fail.

complete(Phrase,AnfangPos,EndePos,PNr) :-
   rule(Neu,[P],RegelNr),           % for any unary rule
   match(P,Phrase,_),               % that matches phrase
   store(chart(Neu,AnfangPos,EndePos,RegelNr,PNr,0,NNr)),
   complete(Neu,AnfangPos,EndePos,NNr),
   fail.

complete(_,_,_,_).

apply_rule(ChartNr,RegelNr) :-
   rule(Neu,[L,R],RegelNr),
   chart(Rechts,MittePos,_,_,_,_,ChartNr),
   match(R,Rechts,_),               % that matches phrase on right
   chart(Links,_,MittePos,_,_,_,_),  
   match(L,Links,_),                % that has left half
   compute_new(L,R,Neu,Links,Ergebnis),
   write(Ergebnis).

apply_rule(ChartNr,RegelNr) :-
   rule(Neu,[P],RegelNr),           % for any unary rule
   chart(Phrase,_,_,_,_,_,ChartNr),
   match(P,Phrase,_),               % that matches phrase
   write(Neu).

%% chart is: [Struktur, AnfangPos, EndPos,
%%                 Mutterregel, Links, Rechts, Nummer]

:- unknown(_,fail).

clear_chart :- abolish(chart,7),
  abolish(chart_counter,1),
  assert(chart_counter(1)).

store(chart(S,A,E,Reg,L,R,N)) :-
   \+ (chart(S,A,E,Reg,L,R,N)),     % don't store same chart entry twice
   chart_entry_number(N),
   write(chart(S,A,E,Reg,L,R,N)),nl,
   assert(chart(S,A,E,Reg,L,R,N)).

chart_entry_number(X) :-
  chart_counter(X),
  abolish(chart_counter,1),
  Y is X+1,
  assert(chart_counter(Y)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% learn new rules
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% adjunct rules

compile_adj :-
   adj(A,B,C,D),
   compile_adj(A,B,C,D),
   fail.

compile_adj.

compile_adj(_,_,_,[]).

compile_adj(Seite,Kat,Merkmale,[Adjunkt|Rest]) :-
   g_translate(Merkmale,Struktur),
   assert(adjunct(Seite,Kat,Struktur,Adjunkt)),
   compile_adj(Seite,Kat,Merkmale,Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% argument rules

compile_arg :-
   arg(A,B,C,D),
   compile_arg(A,B,C,D),
   fail.

compile_arg.

compile_arg(_,_,_,[]).

compile_arg(Seite,Kat,Merkmale,[Adjunkt|Rest]) :-
   g_translate(Merkmale,Struktur),
   assert(argument(Seite,Kat,Struktur,Adjunkt)),
   compile_arg(Seite,Kat,Merkmale,Rest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% composition rules

compile_comp :-
   lc(X,Y),
   compile_lc(X,Y),
   fail.

compile_comp.

compile_lc([LinksKat,LinksML],
           [[Rechts1Kat,Rechts1ML],
            [Rechts2Kat,Rechts2ML]]) :-
   g_t_grammar([LinksKat,LinksML],LinksStruktur),
   g_t_grammar_rec([[Rechts1Kat,Rechts1ML],
                    [Rechts2Kat,Rechts2ML]],
                   RechtsStruktur),
   assert(comp(LinksStruktur,RechtsStruktur)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% transformation rules

compile_transf :-
   tf(X,Y),
   compile_tf(X,Y),
   fail.

compile_transf.

compile_tf([LinksKat,LinksML],
           [[RechtsKat,RechtsML]]) :-
   g_t_grammar([LinksKat,LinksML],LinksStruktur),
   g_t_grammar_rec([[RechtsKat,RechtsML]],RechtsStruktur),
   assert(transf(LinksStruktur,RechtsStruktur)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% learning

% learn rules for one sentence: the main frame routine

learn(SatzNr) :-
   input(SatzNr,Satz),
   write(input(Satz)),nl, 
   parse(SatzNr),
   abolish(parse_counter,1),
   assert(parse_counter(0)),
   parse_count,
   parse_counter(ParseAnzahl),
   kick_off_learning(SatzNr,ParseAnzahl,AnfangsRZ),
   report_learned_rules,
   report_initial_parses(SatzNr,AnfangsRZ),
   unix(system(copy_learned_inf)).
%  assert_rules(AnfangsRZ),
%  final_parsing(SatzNr).

kick_off_learning(SatzNr,0,RZ) :- !,
   abolish(adjunct,4),
   compile_adj,
   abolish(argument,4),
   compile_arg,
   abolish(comp,2),
   compile_comp,
   abolish(transf,2),
   compile_transf,
   learn_rule_counter,
   rule_counter(RZ),
   learn_cycle(SatzNr,0,RZ,RZ).

kick_off_learning(_,_,_) :-
   write('Satz bereits akzeptiert.'),nl.

learn_cycle(SatzNr,0,UrRZ,AltRZ) :- !,
   learn_cycle_learning,
   learn_cycle_parsing(SatzNr,UrRZ,RZ,ParseAnzahl),
   another_learn_cycle(SatzNr,ParseAnzahl,UrRZ,AltRZ,RZ).

learn_cycle(_,_,_,_).

learn_cycle_learning :-
   tell('learn.log'),
   learn_step(1),
   told.

learn_cycle_parsing(SatzNr,UrRZ,RZ,ParseAnzahl) :-
   tell('chart.learn.log'),
   both_f_lc_parsing(SatzNr),
   parse_counter(ParseAnzahl),
   rule_counter(RZ),
   RegelAnzahl is RZ-UrRZ,
   write(RegelAnzahl),write(' neue Regeln, '),
   write(ParseAnzahl),write(' Parses.'),nl.

final_parsing(SatzNr) :-
   tell('chart.log'),
   both_f_lc_parsing(SatzNr),
   parse_counter(ParseAnzahl),
   write(ParseAnzahl),write(' Parses.'),nl.

both_f_lc_parsing(SatzNr) :-
   clear_chart,
   input(SatzNr,Satz),
   write(input(Satz)),nl, 
   process(0),
   told, tell(user),
   abolish(parse_counter,1),
   assert(parse_counter(0)),
   parse_count.

another_learn_cycle(_,_,_,RZ,RZ) :-
   write('Nicht erlernbar.'),nl.

another_learn_cycle(SatzNr,ParseAnzahl,UrRZ,_,RZ) :-
   learn_cycle(SatzNr,ParseAnzahl,UrRZ,RZ).

learn_step(Endposition) :-
   length(Laenge),
   Endposition is Laenge+1.

learn_step(LetztePosition) :-
   length(LetztePosition),
   chart(Rechts,_,LetztePosition,_,_,_,Nr),
   write(LetztePosition),write(': '),write(Nr),nl,
   write('@transformation'),nl,
   learn_transformation(Rechts),
   fail.

learn_step(Position) :-
   length(Laenge),
   Position<Laenge,
   chart(Rechts1,_,Position,_,_,_,LNr),
   chart(Rechts2,Position,_,_,_,_,RNr),
   write(Position),write(': '),
   write(LNr),write('+'),write(RNr),nl,
   write('@adjunct'),nl,
   learn_adjunct(Rechts1,Rechts2),
   write('@argument'),nl,
   learn_argument(Rechts1,Rechts2),
   write('@composition'),nl,
   learn_composition(Rechts1,Rechts2),
   write('@transformation'),nl,
   learn_transformation(Rechts1),
   fail.

learn_step(Position) :-
   NaechstePosition is Position+1,
   learn_step(NaechstePosition).

%%% adjunct rules

learn_adjunct([Kat1,Struktur1],[Kat2,Struktur2]) :-
   adjunct(right,Kat1,Struktur1,Kat2),
   write('@adjunct attempt: '),
   write(adjunct(right,Kat1,Struktur1,Kat2)),nl,
   construct_structure(Kat1,Struktur1,Rechts1S,LinksS),
   construct_structure(Kat2,Struktur2,Rechts2S,_),
   write(LinksS),nl,write(Rechts1S),nl,write(Rechts2S),nl,
   construct_sem_feature(LinksS,Rechts1S,Rechts2S,Nummer,Kat1),
   store_rule([Kat1,LinksS],[[Kat1,Rechts1S],[Kat2,Rechts2S]],Nummer),
   fail.

learn_adjunct([Kat1,Struktur1],[Kat2,Struktur2]) :-
   adjunct(left,Kat2,Struktur2,Kat1),
   write('@adjunct attempt: '),
   write(adjunct(left,Kat2,Struktur2,Kat1)),nl,
   construct_structure(Kat1,Struktur1,Rechts1S,_),
   construct_structure(Kat2,Struktur2,Rechts2S,LinksS),
   write(Rechts2S),nl,write(Rechts1S),nl,write(Rechts2S),nl,
   construct_sem_feature(LinksS,Rechts1S,Rechts2S,Nummer,Kat1),
   store_rule([Kat2,LinksS],[[Kat1,Rechts1S],[Kat2,Rechts2S]],Nummer),
   fail.

learn_adjunct(_,_).

%%% argument rules

learn_argument([Kat1,Struktur1],[Kat2,Struktur2]) :-
   argument(right,Kat1,Struktur1,Kat2),
   write('@argument attempt: '),
   write(argument(right,Kat1,Kat2)),nl,
   write(Struktur1),nl,write(Struktur2),nl,
   req_feature(Kat1,Merkmale1),
   req_feature(Kat2,Merkmale2),
   agreement_check(Merkmale1,Struktur1,Merkmale2,Struktur2),
   write('@passed agreement check'),nl,
   construct_arg_structure(Merkmale1,Merkmale2,Struktur1,Struktur2,
                           Rechts1S,Rechts2S,LinksS),
   write(LinksS),nl,write(Rechts1S),nl,write(Rechts2S),nl,
   write('@enter sem feature construction'),nl,
   construct_sem_feature(LinksS,Rechts1S,Rechts2S,Nummer,Kat1),
   write('@passed sem feature construction'),nl,
   store_rule([Kat1,LinksS],[[Kat1,Rechts1S],[Kat2,Rechts2S]],Nummer),
   fail.

learn_argument([Kat1,Struktur1],[Kat2,Struktur2]) :-
   argument(left,Kat2,Struktur2,Kat1),
   write('@argument attempt: '),
   write(argument(left,Kat1,Kat2)),nl,
   write(Struktur1),nl,write(Struktur2),nl,
   req_feature(Kat1,Merkmale1),
   req_feature(Kat2,Merkmale2),
   agreement_check(Merkmale1,Struktur1,Merkmale2,Struktur2),
   write('@passed agreement check'),nl,
   construct_arg_structure(Merkmale2,Merkmale1,Struktur2,Struktur1,
                           Rechts2S,Rechts1S,LinksS),
   write(LinksS),nl,write(Rechts1S),nl,write(Rechts2S),nl,
   construct_sem_feature(LinksS,Rechts1S,Rechts2S,Nummer,Kat1),
   store_rule([Kat2,LinksS],[[Kat1,Rechts1S],[Kat2,Rechts2S]],Nummer),
   fail.

learn_argument(_,_).

%%% composition rules

learn_composition([R1Kat,R1S],[R2Kat,R2S]) :-
   comp([LKat,LS],[[R1Kat,R1S],[R2Kat,R2S]]),
   strip_sem_feature(LS,LSNeu),
   strip_sem_feature(R1S,R1SNeu),
   strip_sem_feature(R2S,R2SNeu),
   construct_sem_feature(LSNeu,R1SNeu,R2SNeu,Nummer,LKat),
   store_rule([LKat,LSNeu],[[R1Kat,R1SNeu],[R2Kat,R2SNeu]],Nummer),
   fail.

learn_composition(_,_).

%%% transformation rules

learn_transformation([Kat,Struktur]) :-   
   transf([LKat,LS],[[Kat,Struktur]]),
   strip_sem_feature(LS,LSNeu),
   strip_sem_feature(Struktur,StrukturNeu),
   construct_sem_feature_unary(LSNeu,StrukturNeu,Nummer,LKat),
   store_rule([LKat,LSNeu],[[Kat,StrukturNeu]],Nummer),
   fail.

learn_transformation(_).

%%% subs for construction the feature structures in rules

construct_structure(Kat,Quelle,Kopie,ZweitKopie) :-
   req_feature(Kat,Merkmale),
   construct_structure_rec(Kat,Quelle,Kopie,ZweitKopie,Merkmale).

construct_structure_rec(_,_,_,_,[]).

construct_structure_rec(Kat,Quelle,Kopie,ZweitKopie,[Merkmal|Rest]) :-
   write(Merkmal),write(': '),
   g_schema(Merkmal:Wert,Quelle),
   feature_pass(Merkmal,Wert,LernWert),
   g_schema(Merkmal:LernWert,Kopie), 
   g_schema(Merkmal:LernWert,ZweitKopie), 
   construct_structure_rec(Kat,Quelle,Kopie,ZweitKopie,Rest).

   % 1 is head component, 2 is argument
   % feature just in head comp
construct_arg_structure([Merkmal|M1],M2,Q1,Q2,Z1,Z2,ZL) :-
   is_not_in_list(Merkmal,M2),!,
   g_schema(Merkmal:Wert,Q1),
   feature_pass(Merkmal,Wert,LernWert),
   g_schema(Merkmal:LernWert,Z1),
   g_schema(Merkmal:LernWert,ZL),
   construct_arg_structure(M1,M2,Q1,Q2,Z1,Z2,ZL).

   % feature in both arg and head comp
construct_arg_structure([Merkmal|M1],M2,Q1,Q2,Z1,Z2,ZL) :-
   feature_pass(Merkmal,cat,MermalsTyp),
   arg_feature_pass(MermalsTyp,Merkmal,Q1,Q2,Z1,Z2,ZL),
   remove_from_list(Merkmal,M2,M2Neu),
   construct_arg_structure(M1,M2Neu,Q1,Q2,Z1,Z2,ZL).

   % feature just in arg
construct_arg_structure([],[Merkmal|M2],Q1,Q2,Z1,Z2,ZL) :-
   g_schema(Merkmal:Wert,Q1),
   feature_pass(Merkmal,Wert,LernWert),
   g_schema(Merkmal:LernWert,Z2),
   construct_arg_structure([],M2,Q1,Q2,Z1,Z2,ZL).

construct_arg_structure([],[],_,_,_,_,_).

   % agreement features are enforced
arg_feature_pass(agr,Merkmal,_,_,Z1,Z2,ZL) :- !,
   g_schema(Merkmal:Wert,Z1),
   g_schema(Merkmal:Wert,Z2),
   g_schema(Merkmal:Wert,ZL).

   % categorizing features are copied
arg_feature_pass(cat,Merkmal,Q1,Q2,Z1,Z2,ZL) :-
   g_schema(Merkmal:Wert1,Q1),
   g_schema(Merkmal:Wert2,Q2),
   g_schema(Merkmal:Wert1,Z1),
   g_schema(Merkmal:Wert2,Z2),
   g_schema(Merkmal:Wert1,ZL).

   % remove an element from a list
remove_from_list(_,[],[]).

remove_from_list(Element,[Element|Liste],ListeNeu) :- !,
   remove_from_list(Element,Liste,ListeNeu).

remove_from_list(Element,[E2|Liste],[E2|ListeNeu]) :-
   remove_from_list(Element,Liste,ListeNeu).

   % check if agreement features match
agreement_check([Merkmal|M1],Struktur1,M2,Struktur2) :-
   is_not_in_list(Merkmal,M2),!,
   agreement_check(M1,Struktur1,M2,Struktur2).

agreement_check([Merkmal|M1],Struktur1,M2,Struktur2) :-
   (Merkmal=vagr;Merkmal=tagr),!,
   g_schema(Merkmal:Wert1,Struktur1),
   g_schema(Merkmal:Wert2,Struktur2),
   match_plus(Wert1,Wert2,_),
   agreement_check(M1,Struktur1,M2,Struktur2).

agreement_check([Merkmal|M1],Struktur1,M2,Struktur2) :-
   feature_pass(Merkmal,is,agr),!,
   g_schema(Merkmal:Wert,Struktur1),
   g_schema(Merkmal:Wert,Struktur2),
   agreement_check(M1,Struktur1,M2,Struktur2).

agreement_check([_|M1],Struktur1,M2,Struktur2) :-
   agreement_check(M1,Struktur1,M2,Struktur2).

agreement_check([],_,_,_).

   % treat sem feature
strip_sem_feature([A,B,C,D,E,F,G,H,I,J,K,L,M,_],
                  [A,B,C,D,E,F,G,H,I,J,K,L,M,_]).

construct_sem_feature([_,_,_,_,_,_,_,_,_,_,_,_,_,[Nummer,Kat,N1,N2]],
                      [_,_,_,_,_,_,_,_,_,_,_,_,_,N1],
                      [_,_,_,_,_,_,_,_,_,_,_,_,_,N2],
                      Nummer,  Kat).

construct_sem_feature_unary([_,_,_,_,_,_,_,_,_,_,_,_,_,[Nummer,Kat,N1]],
                            [_,_,_,_,_,_,_,_,_,_,_,_,_,N1],
                            Nummer,  Kat).

%%% add a new rule to grammar

store_rule([LK,LS],Rechts,Nummer) :-
   strip_sem_feature(LS,LSohne),
   \+ (rule([LK,L],Rechts,_),        % don't store same rule twice
       strip_sem_feature(L,LSohne)), % will get messy with subsumption!
   rule_number(Nummer),
   write(rule([LK,LS],Rechts,Nummer)),nl,
   assert(rule([LK,LS],Rechts,Nummer)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% report initial parses

report_unlearned_parses(SatzNr) :-
   tell('parses.inf'),
   write(SatzNr),nl,
   chart_counter(ChartEintraege),
   write(ChartEintraege),nl,
   length(Laenge),
   write(Laenge),nl,
   report_p,
   told. 

report_initial_parses(SatzNr,NeuRegelNr) :-
   tell('parses.learn.inf'),
   write(SatzNr),nl,
   write(NeuRegelNr),nl,
   chart_counter(ChartEintraege),
   write(ChartEintraege),nl,
   report_p,
   told.

report_p :-
   length(Laenge),
   chart([s,[_,_,_,_,_,_,_,_,_,_,_,_,_,Sem]],0,Laenge,_,_,_,_),
   write(Sem),nl,
   fail.
   
report_p.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% assert learned rules

assert_rules(2749585).

assert_rules(NeuRegelNr) :-
   abolish(parse,3),
   score_rule_applications(NeuRegelNr),
   find_lowest_score,
   find_good_rules,
   delete_bad_rules(NeuRegelNr).

%%% find rule applications in parse tree

score_rule_applications(NeuRegelNr) :-
   length(Laenge),
   chart([s,_],0,Laenge,_,_,_,ParseNr),
   construct_tree(ParseNr,ParseBaum),
   find_rappl(NeuRegelNr,ParseBaum,Regeln),
   length(Regeln,Anzahl),
   assert(parse(ParseNr,Anzahl,Regeln)),
   write(parse(ParseNr,Anzahl,Regeln)),nl,
   fail.

score_rule_applications(_).

find_rappl(_,[_,_],[]).   % word

find_rappl(NRN,[RegelNr,_,Ast],NeueListe) :-
   find_rappl(NRN,Ast,Liste),
   add_if_new(NRN,RegelNr,Liste,NeueListe).

find_rappl(NRN,[RegelNr,_,Ast1,Ast2],NeueListe) :-
   find_rappl(NRN,Ast1,Liste1),
   find_rappl(NRN,Ast2,Liste2),
   join_new(NRN,Liste1,Liste2,Liste),
   add_if_new(NRN,RegelNr,Liste,NeueListe).

%%% subroutines 

join_new(_,Liste,[],Liste).

join_new(NRN,Liste1,[Element|RestListe2],Liste) :-
   add_if_new(NRN,Element,Liste1,NeueListe),
   join_new(NRN,NeueListe,RestListe2,Liste).
  
add_if_new(NRN,RegelNr,AlteListe,[RegelNr|AlteListe]) :- 
   RegelNr>=NRN,
   is_not_in_list(RegelNr,AlteListe),!.
   
add_if_new(_,_,Liste,Liste).
   
is_not_in_list(_,[]).

is_not_in_list(Element,[Element|_]) :- !, fail.
is_not_in_list(Element,[_|Liste]) :- 
   is_not_in_list(Element,Liste).

%%% find lowest score

find_lowest_score :-
   abolish(lowest_score,1),
   assert(lowest_score(46101)),
   find_lscore.

find_lscore :-
   parse(_,Bewertung,_),
   lowest_score(Niedrigste),
   Bewertung<Niedrigste,
   abolish(lowest_score,1),
   assert(lowest_score(Bewertung)),
   fail.

find_lscore.

%%% find the rules in parse(s) with the lowest score

find_good_rules :-
   abolish(good_rules,1),
   assert(good_rules([])),
   find_grules.

find_grules :-
   lowest_score(Bewertung),
   parse(_,Bewertung,Regeln),
   good_rules(Liste),
   join_new(0,Liste,Regeln,NeueListe),
   abolish(good_rules,1),
   assert(good_rules(NeueListe)),
   write(good_rules(NeueListe)),nl,
   fail.

find_grules.

%%% remove new rules from grammar that are not used in best parses

delete_bad_rules(NRN) :-
   abolish(copy_rule,3),
   copy_old_rules(NRN),
   good_rules(Regeln),
   length(Regeln,Anzahl),
   write(Anzahl),write(' Regeln uebernommen, '),
   copy_good_rules(Regeln),
   abolish(rule,3),
   rewrite_rules,
   abolish(copy_rule,3).

copy_old_rules(NRN) :-
   rule(LS,RS,Nr),
   copy_old_rules_or_not(NRN,LS,RS,Nr),
   fail.

copy_old_rules(_).

copy_old_rules_or_not(NRN,LS,RS,Nr) :-
   Nr<NRN,!,
   assert(copy_rule(LS,RS,Nr)).

copy_old_rules_or_not(_,_,_,_).

copy_good_rules([]).

copy_good_rules([Regel|Rest]) :-
   rule(LS,RS,Regel),
   assert(copy_rule(LS,RS,Regel)),
   copy_good_rules(Rest).

rewrite_rules :-
   copy_rule(LS,RS,Nr),
   assert(rule(LS,RS,Nr)),
   fail.

rewrite_rules.

