#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#define RREQ_FEATURE 8
#define TAGR_FEATURE 10
#define VAGR_FEATURE 11
#define SEMCAT_FEATURE 12
#define SEM_FEATURE 13

#define TEXTWIDTH 160

struct FS
{
  char       name[20][20];
  struct FS  *sub[20];
};

struct FEATURE
{
  int             size;       /* number of features */
  int             text[20];   /* pointer to text entry */
  struct FEATURE  *sub[20];   /* subfeature reference */
};
 
struct MANAGEMENT
{
  int       varcount;         /* number of variables */
  int       var[100];         /* name of variables */ 
  int       ref[100];         /* references to variable */
  int       textcount;        /* position in textbuffer */
  char      text[1000];       /* textbuffer */
};

struct SENTENCE
{
  char  *word[100];
};

struct CHART
{
  int   nr;
  int   length;
  int   rule[10000];
  int   left[10000];
  int   right[10000];
  char *chomsky[10000];
  int   startpos[10000];
  int   endpos[10000];
  int   height[10000];
  int   marked[10000];
};

char read_char(FILE *);
void pretty_line(FILE *,FILE *,char,int,struct CHART *,struct SENTENCE *);
void pretty_title(FILE *,FILE *, char,struct SENTENCE *);
void pretty_section(FILE *,FILE *, char);
void pretty_input(FILE *,FILE *,struct SENTENCE *);
void pretty_chart(FILE *,FILE *, char,int, struct CHART *);
void pretty_rule(FILE *,FILE *, char,int);
void pretty_charttree(FILE *, struct CHART *,struct SENTENCE *);
void pretty_parsetrees(FILE *,struct CHART *,struct SENTENCE *); 
char * read_chomsky(FILE *);
struct FEATURE * read_feature(FILE *,struct MANAGEMENT *);
struct FEATURE * read_vagr(FILE *,struct MANAGEMENT *);
int read_var(FILE *,struct MANAGEMENT *, char *);
int read_number(FILE *,char *);
int read_text(FILE *,struct MANAGEMENT *, char *,int);
void write_feature(FILE *,struct MANAGEMENT *,struct FEATURE *,struct FS *,int);
void mark_chart(struct CHART *,int);
void slope(int,int,int *,int *,float *);

int main(int argc,char *argv[])
{
  FILE *foutput, *finput;
  FILE *countfile;
  struct CHART *chart;
  int varmode=0,rules=0;
  int nr;
  int argused=0;
  char c;
  struct SENTENCE *sentence;

  sentence = malloc(sizeof(struct SENTENCE));
  chart = malloc(sizeof(struct CHART));
  chart->nr = 0;

  foutput=fopen("p.tex","w");
  printf("######%d#######",argc);

  if (argc>1 && argv[1][0]=='r')
    { finput=fopen("rules.inf","r");
      argused++;
      rules=1;}
  else if (argc>1 && argv[1][0]=='l')
    { finput=fopen("rules.learn.inf","r");
      argused++;
      rules=1;}
  else if (argc>1 && argv[1][0]=='c')
    { finput=fopen("chart.learn.log","r");
      argused++;}
  else
    finput=fopen("chart.log","r");  

  countfile=fopen("pretty.inf","r");
  do
  { do 
      c=getc(countfile);
    while(c!='#' && c!=EOF);
    if (c=='#')
      nr=read_number(countfile,&c);
  } while(c!=EOF);
  fclose(countfile);

  countfile=fopen("pretty.inf","a");
  if (argc>2 && argv[1][argused]=='t')
    fprintf(countfile,"#%d: %s\n",nr+1,argv[2]);
  else
    fprintf(countfile,"#%d -\n",nr+1);
  fclose(countfile);

  if (rules==0) 
    fprintf(foutput,"\\twocolumn\n");
  fprintf(foutput,"\\begin{document}\n");
  fprintf(foutput,"\\begin{center}\n");
  if (rules==0)
    fprintf(foutput,"\\huge \\sf Chart Eintr\\%cage \\\\[4mm]\n",34);
  else
    fprintf(foutput,"\\huge \\sf Regeln \\\\[4mm]\n");
  fprintf(foutput,"\\normalsize \\sc Philipp K\\%cohn --- ",34);
  fprintf(foutput,"No. %d --- \\today \\\\",nr+1);
  if (argc>2 && argv[1][argused]=='t')
    fprintf(foutput,"[2mm]\n%s\\\\",argv[2]);
  fprintf(foutput,"[5mm]\n\\small\n");

  c=read_char(finput);
  while (c!=EOF)
  { printf("<%c|%d>",c,c);
    pretty_line(finput,foutput,c,varmode,chart,sentence);
    c=read_char(finput); };
 
  if (rules==0)
  {
    pretty_charttree(foutput,chart,sentence);
    pretty_parsetrees(foutput,chart,sentence);
  }
  fprintf(foutput,"\\end{center}\n\\end{document}");
  fclose(foutput);
  fclose(finput);
}

char read_char(FILE *finput)
{
  char c;

  do
    c=getc(finput);
  while (c==10 || c==13 || c==32);

  return c;
}

void pretty_line(FILE *finput,FILE *foutput,char c,int varmode,
                 struct CHART *chart,struct SENTENCE *sentence)
{
  printf("called pretty line\n");
  if (c=='=')
    pretty_title(finput,foutput,c,sentence);
  else if (c=='S' || c=='C')
    pretty_section(finput,foutput,c);
  else if (c=='i')
    pretty_input(finput,foutput,sentence);
  else if (c=='c')
    pretty_chart(finput,foutput,c,varmode,chart);
  else if (c=='[')
    pretty_rule(finput,foutput,c,varmode);
  else
    printf("Unbekanntes Zeilenanfangssymbol: %c\n",c);
}

void pretty_title(FILE *finput,FILE *foutput,char c,
                  struct SENTENCE *sentence)
{
  int position=0;

  printf("called pretty title\n");
  fprintf(foutput,"\\titel{Position ");

  while (c<'0' || c>'9')
    c=read_char(finput);

  while (c>='0' && c<='9')
  { fprintf(foutput,"%c",c);
    position=position*10+(c-'0');
    c=read_char(finput);};

  fprintf(foutput,"}\n\n");
  fprintf(foutput,"{\\tt %s}\\\\[3mm]\n",sentence->word[position]);

  read_char(finput);
  c=read_char(finput);
  printf("%c",c);
} 

void pretty_input(FILE *finput,FILE *foutput,struct SENTENCE *sentence)
{
  char c,*word;
  int w,i;

  printf("called pretty input\n");
  fprintf(foutput,"{\\tt ");
  
  while (c!='[')
    c=read_char(finput);

  word = calloc(30,sizeof(char));
  i=0;
  w=0;

  while (c!=']')
  { c=read_char(finput);
    if (c==',')
    { fprintf(foutput," ");
      word[i]=0;
      sentence->word[w++]=word;
      i=0;
      word=calloc(30,sizeof(char)); }
    else if (c!=']')
    { fprintf(foutput,"%c",c);
      word[i++]=c; 
  } };

  sentence->word[w]=word;

  fprintf(foutput,"}\n\n");

  read_char(finput);
}

void pretty_section(FILE *finput,FILE *foutput, char c)
{
  printf("called pretty section\n");
  fprintf(foutput,"%cabschnitt{",92);
  
  while (c!='.')
  { fprintf(foutput,"%c",c);
    c=read_char(finput);
  };

  fprintf(foutput,"}\n\n");

  read_char(finput);
  read_char(finput);
}

void pretty_chart(FILE *finput,FILE *foutput, char c,int varmode,
                  struct CHART *chart)
{
  struct MANAGEMENT database;
  struct FS fs,agr,vagr,tagr,rreq;
  char *chomsky;
  struct FEATURE *feature;
  int i;
  int startpos,endpos;
  int nummer,regel,links,rechts;
 
  printf("called pretty chart\n");
  strcpy(fs.name[0],"subcat");
  strcpy(fs.name[1],"case");
  strcpy(fs.name[2],"person");
  strcpy(fs.name[3],"agr");
  strcpy(fs.name[4],"pos");
  strcpy(fs.name[5],"vtyp");
  strcpy(fs.name[6],"det");
  strcpy(fs.name[7],"def");
  strcpy(fs.name[8],"rreq");
  strcpy(fs.name[9],"ragr");
  strcpy(fs.name[10],"tagr");
  strcpy(fs.name[11],"vagr");
  strcpy(fs.name[12],"semcat");
  strcpy(fs.name[13],"sem");
  strcpy(agr.name[0],"number");
  strcpy(agr.name[1],"gender");
  strcpy(vagr.name[0],"subj");
  strcpy(vagr.name[1],"obj1");
  strcpy(vagr.name[2],"obj2");
  strcpy(vagr.name[3],"vcomp");
  strcpy(vagr.name[4],"scomp");
  strcpy(vagr.name[5],"ppobj");
  strcpy(tagr.name[0],"day");
  strcpy(tagr.name[1],"tod");
  strcpy(tagr.name[2],"time");
  strcpy(rreq.name[0],"req");

  fs.sub[3]=&agr;
  fs.sub[8]=&rreq;
  fs.sub[9]=&agr;   /* ragr */
  fs.sub[10]=&tagr;
  fs.sub[11]=&vagr;

  database.varcount=1;
  database.textcount=1;

  read_char(finput); /* h */
  read_char(finput); /* a */
  read_char(finput); /* r */
  read_char(finput); /* t */
  read_char(finput); /* ( */
  read_char(finput); /* [ */

  chomsky=read_chomsky(finput);
  chart->chomsky[chart->nr]=chomsky;

  read_char(finput); /* [ */
  feature=read_feature(finput,&database);

  read_char(finput); /* ] */
  read_char(finput); /* , */

  startpos=read_number(finput,&c);
  printf("established startpos %d\n",startpos);
  endpos=read_number(finput,&c);
  regel=read_number(finput,&c);
  links=read_number(finput,&c);
  rechts=read_number(finput,&c);
  if (regel==0)
    rechts=0;
  nummer=read_number(finput,&c);
  chart->startpos[chart->nr]=startpos;
  chart->endpos[chart->nr]=endpos;
  if (endpos>chart->length)
    chart->length=endpos;
  chart->rule[chart->nr]=regel;
  chart->left[chart->nr]=links;
  chart->right[chart->nr]=rechts;
  chart->nr++;
  printf("finished reading chart entry number %d\n",nummer);

  /*** now comes the output ***/

  fprintf(foutput,"%d $\\bullet$ ",nummer);
  fprintf(foutput,"%s $\\bullet$ ",chomsky);
  if (regel>0)
  { fprintf(foutput,"regel %d $\\bullet$ aus %d ",
            regel,links);
    if (rechts>0)
      fprintf(foutput,"und %d ",rechts);
  }
  else
    fprintf(foutput,"wort ");
  
  fprintf(foutput,"$\\bullet$ von %d bis %d \\nopagebreak\n\n",
          startpos,endpos);

  fprintf(foutput,"\\begin{displaymath}\n");
  write_feature(foutput,&database,feature,&fs,varmode);
  fprintf(foutput,"\\end{displaymath}\n\n");
}

void pretty_rule(FILE *finput,FILE *foutput, char c,int varmode)
{
  struct MANAGEMENT database;
  struct FS fs,agr,vagr,tagr,rreq;
  char *chomsky[3];
  struct FEATURE *links,*rechts1,*rechts2;
  int i;
  int nummer;
 
  printf("called pretty rule\n");
  strcpy(fs.name[0],"subcat");
  strcpy(fs.name[1],"case");
  strcpy(fs.name[2],"person");
  strcpy(fs.name[3],"agr");
  strcpy(fs.name[4],"pos");
  strcpy(fs.name[5],"vtyp");
  strcpy(fs.name[6],"det");
  strcpy(fs.name[7],"def");
  strcpy(fs.name[8],"rreq");
  strcpy(fs.name[9],"ragr");
  strcpy(fs.name[10],"tagr");
  strcpy(fs.name[11],"vagr");
  strcpy(fs.name[12],"semcat");
  strcpy(fs.name[13],"sem");
  strcpy(agr.name[0],"number");
  strcpy(agr.name[1],"gender");
  strcpy(vagr.name[0],"subj");
  strcpy(vagr.name[1],"obj1");
  strcpy(vagr.name[2],"obj2");
  strcpy(vagr.name[3],"vcomp");
  strcpy(vagr.name[4],"scomp");
  strcpy(vagr.name[5],"ppobj");
  strcpy(tagr.name[0],"day");
  strcpy(tagr.name[1],"tod");
  strcpy(tagr.name[2],"time");
  strcpy(rreq.name[0],"req");

  fs.sub[3]=&agr;
  fs.sub[8]=&rreq;
  fs.sub[9]=&agr;   /* ragr */
  fs.sub[10]=&tagr;
  fs.sub[11]=&vagr;

  database.varcount=1;
  database.textcount=1;

  read_char(finput); /* [ */

  chomsky[0]=read_chomsky(finput);

  read_char(finput); /* [ */
  links=read_feature(finput,&database);

  read_char(finput); /* ] */
  read_char(finput); /* , */

  read_char(finput); /* [ of the right side list */
  c=read_char(finput);
  if (c=='[')
  { chomsky[1]=read_chomsky(finput);
    read_char(finput);
    rechts1=read_feature(finput,&database);
    c=read_char(finput);
    if (c==',')
    { read_char(finput);
      chomsky[2]=read_chomsky(finput);
      read_char(finput);
      rechts2=read_feature(finput,&database);
      read_char(finput); } /* ] of the right side list */
    else
      rechts2=NULL;
  }
  else 
    rechts1=NULL;  

  read_char(finput); /* , */
  nummer=read_number(finput,&c);
  printf("finished reading rule number %d\n",nummer);

  /*** now comes the output ***/

  fprintf(foutput,"%d $\\bullet$ {\\tt ",nummer);
  fprintf(foutput,"%s $\\rightarrow$ ",chomsky[0]);
  if (rechts1!=NULL)
  { fprintf(foutput,"%s ",chomsky[1]);
    if (rechts2!=NULL)
      fprintf(foutput,"%s ",chomsky[2]); }
  fprintf(foutput,"} \\nopagebreak\n\n");

  fprintf(foutput,"\\begin{displaymath}\n");
  write_feature(foutput,&database,links,&fs,varmode);
  fprintf(foutput,"\\rightarrow \n");
  if (rechts1!=NULL)
  { write_feature(foutput,&database,rechts1,&fs,varmode);
    if (rechts2!=NULL)
      write_feature(foutput,&database,rechts2,&fs,varmode); } 
  fprintf(foutput,"\\end{displaymath}\n\n");
  fprintf(foutput,"$\\rule{10cm}{0.5pt}$\n\n");
}

char * read_chomsky(FILE *finput)
{
  int i=0;
  char c,*kat;

  kat=malloc(sizeof(char)*10);

  printf("called read_chomsky:");
  c=read_char(finput); 
  while (c!=',')
  { kat[i++] = c;
    printf("%c",c);
    c=read_char(finput); }
  kat[i]=0;
  printf("\n");
  return kat;
}

struct FEATURE * read_vagr(FILE *finput,struct MANAGEMENT *database)
{
  int i=0;
  char c;
  struct FEATURE *feature;
  feature=malloc(sizeof(struct FEATURE));

  printf("called read vagr\n");
  do 
  { c=read_char(finput);
    printf("<<%c>>... ",c);
    if (c=='_')
      feature->text[i]=read_text(finput,database,&c,0);
    else
    { c=read_char(finput);
      feature->text[i]=read_text(finput,database,&c,1);
      c=read_char(finput); }
    i++;
  } while (c==',');

  feature->size = i;
  return feature;
}
struct FEATURE * read_feature(FILE *finput,struct MANAGEMENT *database)
{
  int i=0,klammern;
  char c;
  struct FEATURE *feature;
  feature=malloc(sizeof(struct FEATURE));

  printf("called read feature\n");
  do 
  { c=read_char(finput);
    printf("<<%c>>... ",c);
    if (c=='[')
    { feature->text[i]=0;
      if (i==TAGR_FEATURE || i==VAGR_FEATURE || i==RREQ_FEATURE)
        feature->sub[i]=read_vagr(finput,database); /*[has,...]*/
      else if (i==SEMCAT_FEATURE)
      { feature->text[i]=read_text(finput,database,&c,2); /* pair */
        printf("should still be comma: %c\n",c);}
      else
        feature->sub[i]=read_feature(finput,database); /* subfeat */
      c=read_char(finput); }
    else
      feature->text[i]=read_text(finput,database,&c,0);
    i++;
  } while (c==',' && i!=SEM_FEATURE);
  printf("finished feature with %c at %d\n",c,i);

  if (i==SEM_FEATURE && c==',')
  {
    printf("sem-feature skip over:");
    c=read_char(finput);
    if (c=='[')
    { klammern=1;
      do 
      { c=read_char(finput);
        printf("%c",c);
        if (c=='[')
          klammern++;
        else if (c==']')
          klammern--;
      } while (klammern>0); }
    else
      read_number(finput,&c);
    read_char(finput);      
  }
  feature->size = i;
  return feature;
}

int read_var(FILE *finput,struct MANAGEMENT *database, char *lastchar)
{
  int value,i,number;
  char c;

  printf("called read var\n");
  value=read_number(finput,lastchar);
  if (value!=2749585)
  { printf("found variable _%d ... ",value);

    for (i=1;i<database->varcount;i++)
      if (database->var[i]==value)
      {
        printf("exists as #%d of %d\n",i,database->varcount-1);
        database->ref[i]++;
        number=i;
        i=46101-1;
      }

    if (i!=46101)
    {
      printf("established variable #%d: _%d\n",i,value);
      database->var[i]=value;
      database->ref[i]=1;
      database->varcount++;
      number=i;
    }
  }
  else 
    number=value;
  return number;
}

int read_number(FILE *finput,char *lastchar)
{
  int value=0;
  int vorzeichen=1;
  char c;

  c=read_char(finput);
  if (c=='-')
  { vorzeichen=-1;
    c=read_char(finput); }
  if (c<'0' || c>'9')
    value=2749585;
  else
    while (c>='0' && c<='9')
    {
      value=value*10+(c-'0');
      c=read_char(finput);
    }
  *lastchar=c;
  return (value*vorzeichen);
}

int read_text(FILE *finput,struct MANAGEMENT *database, char *c,
              int specialflag)
{
  int start,i,var,kommaflag=0;
  start=database->textcount;

  printf("called read text\n");
  i=database->textcount;
  if (specialflag==1)
  { database->text[i++]='&';
    printf("special vagr treatment "); }
  else if (specialflag==2)
  { database->text[i++]='(';
    *c=read_char(finput); 
    printf("special pp semcat treatment "); }
  do
  { if (*c=='_')
    { var=read_var(finput,database,c);
      if (var!=2749585)
      { database->text[i++]='#';
        database->text[i++]=var;}
      else
        database->text[i++]='-'; }
    else if (*c=='[' & kommaflag==1)
    { specialflag=3;           /* ppobj semcat in vagr */
      *c=read_char(finput); }
    else
    { database->text[i++]=*c;
      *c=read_char(finput); }
    if (*c==',')
      kommaflag=1;
  } while ((*c!=',' || specialflag!=0) && *c!=']');
  if (specialflag==2)
    database->text[i++]=')';
  if (specialflag==3)
    { *c=read_char(finput);
      printf("should be a comma: %c\n",*c); }
  database->text[i++]=0;
  database->textcount=i;

  return start;
}

void write_feature(FILE *foutput,struct MANAGEMENT *database,
                   struct FEATURE *feature, struct FS *fs,
                   int varmode) 
{
  int i,pos,var;
  char c;

  fprintf(foutput,"\\left[ \\begin{array}{l} ");
  for (i=0; i<feature->size; i++)
  {
    if (feature->text[i] == 0)
    { fprintf(foutput,"{\\sc %s}\\!:\\, ",fs->name[i]);
      write_feature(foutput,database,feature->sub[i],
                    fs->sub[i],varmode);
      fprintf(foutput,"\\\\ \n");
    }
    else
    { 
      pos=feature->text[i];
      c=database->text[pos++];
      if (c=='&')
      { c=database->text[pos++];
        if(c!='#')
        {
          fprintf(foutput,"{\\sc %s}\\!:\\, {\\rm ",fs->name[i]);
          do
          { fprintf(foutput,"%c",c);
            c=database->text[pos++];
          } while (c!=',');
          fprintf(foutput,"} ");
          c=database->text[pos++];
          if (c=='#')
          { var=database->text[pos++];
            if (database->ref[var] > 1) /** only multiple ref. var */
            { if (varmode==0)
                fprintf(foutput,"\\langle %d \\rangle \\\\ ",var);
              else
                fprintf(foutput,"\\langle %d \\rangle \\\\ ",
                                database->var[var]);}
            else
              fprintf(foutput,"\\\\ ");
          }
          else
          { fprintf(foutput,"(");
            do 
            { if (c=='#')
              { var=database->text[pos++];
                if (database->ref[var] > 1) /** multiple ref. var */
                { if (varmode==0)
                    fprintf(foutput,"\\langle %d \\rangle ",var);
                  else
                    fprintf(foutput,"\\langle %d \\rangle ",
                                    database->var[var]);} 
                else
                  fprintf(foutput,"\\langle \\rangle");}
              else
                fprintf(foutput,"%c",c); 
              c=database->text[pos++]; 
            } while (c!=0); 
            fprintf(foutput,")\\\\ \n");
      } } }
      else if(c=='#' && database->text[pos+1]==0)
      { var=database->text[pos++];
        if (database->ref[var] > 1) /** only multiple ref. var */
        { fprintf(foutput,"{\\sc %s}\\!:\\, ",fs->name[i]);
          if (varmode==0)
            fprintf(foutput,"\\langle %d \\rangle \\\\ ",var); 
          else
            fprintf(foutput,"\\langle %d \\rangle \\\\ ",
                    database->var[var]); }
        pos++;
      }
      else
      { fprintf(foutput,"{\\sc %s}\\!:\\, ",fs->name[i]);
        while(c!=0)
        { if (c!='#')
            fprintf(foutput,"%c",c);
          else
          { var=database->text[pos++];
            if (database->ref[var] > 1) /** multiple ref. var */
            { if (varmode==0)
                fprintf(foutput,"\\langle %d \\rangle ",var); 
              else
                fprintf(foutput,"\\langle %d \\rangle ",
                        database->var[var]); }
            else
              fprintf(foutput,"\\langle \\rangle "); }
          c=database->text[pos++];
        }  
        fprintf(foutput,"\\\\ \n");
    } }
  }
  fprintf(foutput,"\\end{array} \\right] ");
}

void pretty_charttree(FILE *foutput,struct CHART *chart, 
                      struct SENTENCE *sentence) 
{
  int nr,i,j,unit,p,pos,l,c;
  int height[1000],h,maxheight;
  int line_start[1000],line_end[1000],line_chart[1000];

  nr=chart->nr;
  l=chart->length;
  unit=TEXTWIDTH/l;   /* Breite der Boxen, max. 18 */
  if (unit>18)
    unit=18;

  for (j=0;j<l;j++)
    height[j]=0;

  for (i=0;i<nr;i++)   /* Hoehenposition der Boxen ermitteln */
  { h=0;
    for (j=chart->startpos[i];j<chart->endpos[i];j++)
      if (height[j]>h)
        h=height[j];
    h++;
    chart->height[i]=h;
    for (j=chart->startpos[i];j<chart->endpos[i];j++)
      height[j]=h;
    if (h>maxheight)
      maxheight=h;
  }

  fprintf(foutput,"\\onecolumn\n");
  fprintf(foutput,"\\huge \\sf Chart \\%cUberblick \\\\[4mm]\n",34);
  fprintf(foutput,"\\footnotesize\n");
  
  for (j=maxheight; j>0; j--)
  { p=0;
    printf("hoehe %d, max. %d\n",j,maxheight);
    for (i=0;i<nr;i++)
      if (chart->height[i]==j)
      { line_chart[p]=i;
        line_start[p]=chart->startpos[i];
        line_end[p++]=chart->endpos[i];
      } 
    pos=0;
    for (i=0;i<p;i++)
    { c=line_chart[i];
      if (pos<line_start[i])
        fprintf(foutput,"\\makebox[%dmm]{\\hfill\n}",
                        unit*(line_start[i]-pos));
      fprintf(foutput,"\\makebox[%dmm]{\\,%d\\hfill ",
                        unit*(line_end[i]-line_start[i]), c+1); 
      if (chart->left[c]==0)
        fprintf(foutput,"w\\hfill");
      else if (chart->right[c]==0)
        fprintf(foutput,"%d(%d)\\hfill",
                        chart->rule[c],chart->left[c]);
      else
        fprintf(foutput,"%d(%d,%d)\\hfill",chart->rule[c],
                        chart->left[c],chart->right[c]);
      fprintf(foutput,"\\tt %s\n}",chart->chomsky[c]);
      pos=line_end[i];
    }
    if (pos<l)
      fprintf(foutput,"\\makebox[%dmm]{\\hfill\n}",unit*(l-pos));
    fprintf(foutput,"\n\\\\[-2.5mm]\n");

    pos=0;
    for (i=0;i<p;i++)
    { c=line_chart[i];
      if (pos<line_start[i])
        fprintf(foutput,"\\makebox[%dmm]{\\hfill\n}",
                        unit*(line_start[i]-pos));
      fprintf(foutput,"\\makebox[%dmm]{\\,\\hrulefill\n}",
                        unit*(line_end[i]-line_start[i]));
      pos=line_end[i];
    }
    if (pos<l)
      fprintf(foutput,"\\makebox[%dmm]{\\hfill\n}",unit*(l-pos));
    fprintf(foutput,"\n\\\\\n");
  } 
  for (i=0;i<l;i++)
    fprintf(foutput,"\\makebox[%dmm]{\\hfill\\em %s\\hfill\n}",
                    unit,sentence->word[i]);
}

void pretty_parsetrees(FILE *foutput,struct CHART *chart,
                       struct SENTENCE *sentence) 
{
  int nr,i,j,unit,p,pos,l,c,uy;
  int x,y; float error;
  int parse[20],pc=0,pnr;
  int height[100],h,maxheight;
  int line_start[100],line_end[100],line_chart[100];

  nr=chart->nr;
  l=chart->length;
  unit=TEXTWIDTH/l;
  if (unit>12)
    unit=12;
  uy=12;

  for (i=0;i<nr;i++)
  { printf("*%d: %d,%d,%d\n",i,strcmp(chart->chomsky[i],"s")==0,
        chart->startpos[i]==0,chart->endpos[i]==l);
    if (strcmp(chart->chomsky[i],"s")==0 && 
        chart->startpos[i]==0 && chart->endpos[i]==l)
      parse[pc++]=i;
  }

 pnr=-1;
 while(++pnr<pc)
 {
  for (i=0;i<nr;i++)
    chart->marked[i]=0;
  printf("parse #%d:%d/n",pnr,parse[pnr]+1);
  fprintf(foutput,"\\newpage \\huge\n");
  fprintf(foutput,"\\sf Parse Tree\\\\[4mm]\n",pnr+1,pc);
  fprintf(foutput,"\\normalsize \\sc %d von %d \\\\[8mm]\n",pnr+1,pc);
  mark_chart(chart,parse[pnr]);

  for (j=0;j<l;j++)
    height[j]=0;

  for (i=0;i<nr;i++)
  if (chart->marked[i]==1)
  { h=0;
    for (j=chart->startpos[i];j<chart->endpos[i];j++)
      if (height[j]>h)
        h=height[j];
    h++;
    chart->height[i]=h;
    for (j=chart->startpos[i];j<chart->endpos[i];j++)
      height[j]=h;
    if (h>maxheight)
      maxheight=h;
  }

  fprintf(foutput,"\\small \\unitlength 1mm \n");
  fprintf(foutput,"\\begin{picture}(%d,%d)(0,0)\n",
                  unit*l,uy*maxheight,unit*l/2,uy*maxheight/2);

  for (c=0;c<nr;c++)
    if (chart->marked[c]==1)
    { fprintf(foutput,"\\put(%d,%d){",
              (chart->endpos[c]+chart->startpos[c])*unit/2-20,
              chart->height[c]*uy);
      fprintf(foutput,"\\parbox[tc]{4cm}{\\center %d: {\\tt %s}\\\\",
                        c+1, chart->chomsky[c]);
      if (chart->left[c]==0)
        fprintf(foutput,"\\em %s",sentence->word[chart->startpos[c]]);
      else
        fprintf(foutput,"rule %d",chart->rule[c]);
      fprintf(foutput,"}}\n");

      if (chart->left[c]>0)
        if (chart->right[c]>0)
        { slope((chart->endpos[chart->left[c]-1]+
                 chart->startpos[chart->left[c]-1])*unit/2-
                (chart->endpos[c]+chart->startpos[c])*unit/2,
                chart->height[chart->left[c]-1]*uy-
                chart->height[c]*uy-4+uy,
                &x,&y,&error);
          if (error==0)
            fprintf(foutput,"\\put(%d,%d){\\line(%d,%d){%d}}\n",
                         (chart->endpos[c]+chart->startpos[c])*unit/2,
                          chart->height[c]*uy-5,   0,-1,
                          chart->height[c]*uy-
                          chart->height[chart->left[c]-1]*uy-6);
          else 
            fprintf(foutput,"\\put(%d,%d){\\line(%d,%d){%f}}\n",
                         (chart->endpos[c]+chart->startpos[c])*unit/2,
                         chart->height[c]*uy-5,
                         x,y,
                         (chart->endpos[chart->left[c]-1]+
                          chart->startpos[chart->left[c]-1]-
                          chart->endpos[c]-chart->startpos[c])
                         *unit/2/error);

          slope((chart->endpos[chart->right[c]-1]+
                 chart->startpos[chart->right[c]-1])*unit/2-
                (chart->endpos[c]+chart->startpos[c])*unit/2,
                chart->height[chart->right[c]-1]*uy-
                chart->height[c]*uy-4+uy,
                &x,&y,&error);
          if (error==0)
            fprintf(foutput,"\\put(%d,%d){\\line(%d,%d){%d}}\n",
                         (chart->endpos[c]+chart->startpos[c])*unit/2,
                          chart->height[c]*uy-5,   0,-1,
                          chart->height[c]*uy-
                          chart->height[chart->right[c]-1]*uy-6);
          else 
            fprintf(foutput,"\\put(%d,%d){\\line(%d,%d){%f}}\n",
                         (chart->endpos[c]+chart->startpos[c])*unit/2,
                         chart->height[c]*uy-5,
                         x,y,
                         (chart->endpos[chart->right[c]-1]+
                          chart->startpos[chart->right[c]-1]-
                          chart->endpos[c]-chart->startpos[c])
                         *unit/2/error);
        }
        else
          fprintf(foutput,"\\put(%d,%d){\\line(%d,%d){%d}}\n",
                         (chart->endpos[c]+chart->startpos[c])*unit/2,
                          chart->height[c]*uy-5,
                          0,-1,uy-8);
                          
    }

  fprintf(foutput,"\\end{picture}\n");
}} 

void mark_chart(struct CHART *chart,int n)
{
  chart->marked[n]=1;
  printf("marked %d: %d %d\n",n+1,chart->left[n],chart->right[n]);
  if (chart->left[n]>0)
    mark_chart(chart,chart->left[n]-1);
  if (chart->right[n]>0)
    mark_chart(chart,chart->right[n]-1);

}

void slope(int x, int y, int *xx, int *yy, float *error)
{
  int i,j;
  float minerror=666;
  printf("\ntrying to correct %d;%d:",x,y);
  for (i=1; i<=6; i++)
    for (j=1; j<=6; j++)
    { *error=(float)j/(float)i / (float)y*(float)x;
      if (*error<0)
        *error=-*error;
      if (*error<1)
        *error=1 / *error;
      printf("\n%d;%d produces %f",i,j,*error);
      if (*error<minerror)
      { minerror=*error;
        *xx=i;
        *yy=j;
        printf(" - taken");
    } }
   *yy = -*yy;
   *error=((float)*yy)/ *xx / ((float)y)* x;
   if (*error<0)
     *xx=-*xx;
}
